# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/4/19 23:27
describe: 
"""
import numpy as np
import pandas as pd


def subtract_fee(df, fee=1):
    """依据单品种持仓信号扣除手续费"""
    assert 'dt' in df.columns, 'dt 列必须存在'
    assert 'pos' in df.columns, 'pos 列必须存在'
    assert all(x in [0, 1, -1] for x in df['pos'].unique()), "pos 列的值必须是 0, 1, -1 中的一个"

    if 'n1b' not in df.columns:
        assert 'price' in df.columns, '当n1b列不存在时，price 列必须存在'
        df['n1b'] = (df['price'].shift(-1) / df['price'] - 1) * 10000
    
    df['date'] = df['dt'].dt.date
    df['edge_pre_fee'] = df['pos'] * df['n1b']
    df['edge_post_fee'] = df['pos'] * df['n1b']

    # 扣费规则, 开仓扣费在第一个持仓K线上，平仓扣费在最后一个持仓K线上
    open_pos = (df['pos'].shift() != df['pos']) & (df['pos'] != 0)
    exit_pos = (df['pos'].shift(-1) != df['pos']) & (df['pos'] != 0)
    df.loc[open_pos, 'edge_post_fee'] = df.loc[open_pos, 'edge_post_fee'] - fee
    df.loc[exit_pos, 'edge_post_fee'] = df.loc[exit_pos, 'edge_post_fee'] - fee
    return df


def daily_performance(daily_returns):
    """计算日收益数据的年化收益率、夏普比率、最大回撤、卡玛比率

    :param daily_returns: 日收益率数据，样例：
        [0.01, 0.02, -0.01, 0.03, 0.02, -0.02, 0.01, -0.01, 0.02, 0.01]
    :return: dict
    """
    if isinstance(daily_returns, list):
        daily_returns = np.array(daily_returns)
    
    if len(daily_returns) == 0 or np.std(daily_returns) == 0 or all(x == 0 for x in daily_returns):
        return {"年化": 0, "夏普": 0, "最大回撤": 0, "卡玛": 0}
    
    annual_returns = (1 + np.sum(daily_returns) / len(daily_returns)) ** 252 - 1
    sharpe_ratio = np.mean(daily_returns) / np.std(daily_returns) * np.sqrt(252)
    cum_returns = np.cumprod(1 + daily_returns)
    max_drawdown = np.max(np.maximum.accumulate(cum_returns) - cum_returns) / np.max(cum_returns)
    kama = annual_returns / max_drawdown if max_drawdown != 0 else 10
    win_pct = len(daily_returns[daily_returns > 0]) / len(daily_returns)
    return {
        "年化": round(annual_returns, 4),
        "夏普": round(sharpe_ratio, 2),
        "回撤": round(max_drawdown, 4),
        "卡玛": round(kama, 2),
        "胜率": round(win_pct, 4),
    }


def net_value_stats(nv: pd.DataFrame, exclude_zero: bool = False, sub_cost=True) -> dict:
    """统计净值曲线的年化收益、夏普等

    :param nv: 净值数据，格式如下：

                           dt  edge  cost
        0 2017-01-03 09:30:00   0.0   0.0
        1 2017-01-03 10:00:00   0.0   0.0
        2 2017-01-03 10:30:00   0.0   0.0
        3 2017-01-03 11:00:00   0.0   0.0
        4 2017-01-03 13:00:00   0.0   0.0

        列说明：
        dt: 交易时间
        edge: 单利收益，单位：BP
        cost: 交易成本，单位：BP；可选列，如果没有成本列，则默认为0

    :param exclude_zero: 是否排除收益为0的情况，一般认为收益为0的情况是没有持仓的
    :param sub_cost: 是否扣除成本
    :return:
    """
    nv = nv.copy(deep=True)
    nv['dt'] = pd.to_datetime(nv['dt'])

    if sub_cost:
        assert 'cost' in nv.columns, "成本列cost不存在"
        nv['edge'] = nv['edge'] - nv['cost']
    else:
        if 'cost' not in nv.columns:
            nv['cost'] = 0

    if exclude_zero:
        nv = nv[(nv['edge'] != 0) | (nv['cost'] != 0)]

    # 按日期聚合
    nv['date'] = nv['dt'].apply(lambda x: x.date())
    df_nav = nv.groupby('date')['edge'].sum() / 10000
    df_nav = df_nav.cumsum()

    if all(x == 0 for x in nv['edge']):
        # 处理没有持仓记录的情况
        sharp = 0
        y_ret = 0
        calmar = 0
        mdd = 0
    else:
        # y_ret = yearly return
        N = 252
        y_ret = df_nav.iloc[-1] * (N / len(df_nav))
        if df_nav.diff().std() != 0:
            sharp = df_nav.diff().mean() / df_nav.diff().std() * pow(N, 0.5)
        else:
            sharp = 0
        df0 = df_nav.shift(1).ffill().fillna(0)
        mdd = (1 - (df0 + 1) / (df0 + 1).cummax()).max()
        calmar = y_ret / mdd if mdd != 0 else 1

    prefix = "有持仓时间" if exclude_zero else ""
    res = {"夏普": round(sharp, 2), "卡玛": round(calmar, 2), "年化": round(y_ret, 4), "最大回撤": round(mdd, 4)}
    res = {f"{prefix}{k}": v for k, v in res.items()}

    if not exclude_zero:
        res['持仓覆盖'] = round(len(nv[(nv['edge'] != 0) | (nv['cost'] != 0)]) / len(nv), 4) if len(nv) > 0 else 0
    return res



from loguru import logger
from czsc import CZSC
from collections import OrderedDict
from czsc.utils import create_single_signal, get_sub_elements


def bar_eight_V230702(c: CZSC, **kwargs) -> OrderedDict:
    """8K走势分类

    参数模板："{freq}_D{di}#8K_走势分类V230702"

    **信号逻辑：**

    参见博客：https://blog.sina.com.cn/s/blog_486e105c010009uy.html
    这篇博客给出了8K走势分类的逻辑。

    **信号列表：**

    - Signal('30分钟_D1#8K_走势分类V230702_弱平衡市_任意_任意_0')
    - Signal('30分钟_D1#8K_走势分类V230702_双中枢下跌_任意_任意_0')
    - Signal('30分钟_D1#8K_走势分类V230702_转折平衡市_任意_任意_0')
    - Signal('30分钟_D1#8K_走势分类V230702_强平衡市_任意_任意_0')
    - Signal('30分钟_D1#8K_走势分类V230702_双中枢上涨_任意_任意_0')
    - Signal('30分钟_D1#8K_走势分类V230702_无中枢上涨_任意_任意_0')
    - Signal('30分钟_D1#8K_走势分类V230702_无中枢下跌_任意_任意_0')

    :param c: CZSC对象
    :return: 信号识别结果
    """
    di = int(kwargs.get("di", 1))
    freq = c.freq.value
    k1, k2, k3 = f"{freq}_D{di}#8K_走势分类V230702".split("_")
    v1 = "其他"
    if len(c.bars_raw) < di + 12:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    bars = get_sub_elements(c.bars_raw, di=di, n=8)
    zs_list = []
    for b1, b2, b3 in zip(bars[:-2], bars[1:-1], bars[2:]):
        if min(b1.high, b2.high, b3.high) > max(b1.low, b2.low, b3.low):
            zs_list.append([b1, b2, b3])
    
    _dir = "上涨" if bars[-1].close > bars[0].open else "下跌"

    if not zs_list:
        v1 = f"无中枢{_dir}"
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)
    
    # 双中枢的情况，有一根K线的 high low 与前后两个中枢没有重叠
    if len(zs_list) >= 2:
        zs1, zs2 = zs_list[0], zs_list[-1]
        zs1_high, zs1_low = max([x.high for x in zs1]), min([x.low for x in zs1])
        zs2_high, zs2_low = max([x.high for x in zs2]), min([x.low for x in zs2])
        if _dir == "上涨" and zs1_high < zs2_low:
            v1 = f"双中枢{_dir}"
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)
        
        if _dir == "下跌" and zs1_low > zs2_high:
            v1 = f"双中枢{_dir}"
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)
        
    # 单中枢的情况，前三根K线出现高点：弱平衡市，前三根K线出现低点：强平衡市，否则：转折平衡市
    high_first = max(bars[0].high, bars[1].high, bars[2].high) == max([x.high for x in bars])
    low_first = min(bars[0].low, bars[1].low, bars[2].low) == min([x.low for x in bars])
    if high_first and not low_first:
        v1 = "弱平衡市"
    elif low_first and not high_first:
        v1 = "强平衡市"
    else:
        v1 = "转折平衡市"
        
    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)



def check():
    from czsc.connectors import research
    from czsc.traders.base import check_signals_acc

    symbols = research.get_symbols('A股主要指数')
    bars = research.get_raw_bars(symbols[0], '15分钟', '20181101', '20210101', fq='前复权')

    signals_config = [{'name': bar_eight_V230702, 'di': 1, 'freq': '30分钟'}]
    check_signals_acc(bars, signals_config=signals_config, height='780px', delta_days=0) # type: ignore


if __name__ == '__main__':
    check()



# -*- coding: utf-8 -*-
import os

os.environ['czsc_max_bi_num'] = '20'
os.environ['czsc_research_cache'] = r"D:\CZSC投研数据"
os.environ['signals_module_name'] = 'czsc.signals'
import json
import streamlit as st
import pandas as pd
from typing import List
from copy import deepcopy
from czsc.utils.bar_generator import freq_end_time
from czsc.connectors.research import get_symbols, get_raw_bars
from czsc import CzscStrategyBase, Position, CzscTrader, KlineChart, Freq, Operate

st.set_page_config(layout="wide")

class Strategy(CzscStrategyBase):
    @property
    def positions(self) -> List[Position]:
        """返回当前的持仓策略"""
        json_strategies = self.kwargs.get("json_strategies")
        assert json_strategies, "请在初始化策略时，传入参数 json_strategies"
        positions = []
        for _, pos in json_strategies.items():
            pos["symbol"] = self.symbol
            positions.append(Position.load(pos))
        return positions

if 'bar_edt_index' not in st.session_state:
    st.session_state.bar_edt_index = 1

if 'run' not in st.session_state:
    st.session_state.run = False

if 'refresh' not in st.session_state:
    st.session_state.refresh = True

def update_bar_edt():
    bar_edt_ = pd.Timestamp(st.session_state.bar_edt)
    i = 1
    for x in bars_right:
        if x.dt >= bar_edt_:
            break
        i += 1
    st.session_state.bar_edt_index = i

def update_change():
    st.session_state.refresh = True
    st.session_state.bar_edt_index = 1

@st.cache_data
def get_bars(symbol_, base_freq_, sdt_, edt_):
    return get_raw_bars(symbol_, base_freq_, sdt=sdt_ - pd.Timedelta(days=365 * 3), edt=edt_)

with st.sidebar:
    st.title("CZSC策略回放工具")
    symbol = st.selectbox("选择标的：", get_symbols('ALL'), index=0, on_change=update_change)
    sdt = st.date_input("开始日期：", value=pd.to_datetime('2022-01-01'), on_change=update_change)
    edt = st.date_input("结束日期：", value=pd.to_datetime('2022-12-31'), on_change=update_change)
    files = st.file_uploader(label='上传策略文件', type='json', accept_multiple_files=True, on_change=update_change)

    if files:
        if st.session_state.refresh:
            json_strategies = {file.name: json.loads(file.getvalue().decode("utf-8")) for file in files}
            tactic: CzscStrategyBase = Strategy(
                symbol=symbol, signals_module_name=os.environ['signals_module_name'], json_strategies=json_strategies
            )
            bars = get_bars(symbol, tactic.base_freq, sdt, edt)
            bg, bars_right = tactic.init_bar_generator(bars, sdt=sdt)
            bars_num = len(bars_right)
            if bars_num > 1000:
                st.warning("回放数据量较大（超过1000根K线），建议缩小回放时间范围")
        else:
            st.session_state.refresh = False

        c1, c2, c3, c4 = st.columns([5, 5, 5, 5])
        with c1:
            button_run = st.button('播放')
            if button_run:
                st.session_state.run = True

        with c2:
            button_run = st.button('暂停')
            if button_run:
                st.session_state.run = False

        with c3:
            button_dec = st.button('左移')
            if button_dec:
                st.session_state.bar_edt_index -= 1
                st.session_state.bar_edt_index = max(st.session_state.bar_edt_index, 0)

        with c4:
            button_inc = st.button('右移')
            if button_inc:
                st.session_state.bar_edt_index += 1
                st.session_state.bar_edt_index = min(st.session_state.bar_edt_index, bars_num)

        bar_edt = bars_right[st.session_state.bar_edt_index - 1].dt
        st.text_input('定位到指定时间', value=bar_edt.strftime('%Y-%m-%d %H:%M'), key="bar_edt", on_change=update_bar_edt)
        st.success(f"共{bars_num}根K线, 时间范围：{bars_right[0].dt} - {bars_right[-1].dt}; 当前K线：{bar_edt}")

def show_trader(trader: CzscTrader):
    freqs = trader.freqs
    tabs = st.tabs(freqs + ['最后信号', '收益分析', '策略详情'])

    i = 0
    for freq in freqs:
        c = trader.kas[freq]
        df = pd.DataFrame(c.bars_raw)
        kline = KlineChart(n_rows=3, title='', width="100%", height=800)
        kline.add_kline(df, name="")

        if len(c.bi_list) > 0:
            bi = pd.DataFrame(
                [{'dt': x.fx_a.dt, "bi": x.fx_a.fx} for x in c.bi_list]
                + [{'dt': c.bi_list[-1].fx_b.dt, "bi": c.bi_list[-1].fx_b.fx}]
            )
            fx = pd.DataFrame([{'dt': x.dt, "fx": x.fx} for x in c.fx_list])
            kline.add_scatter_indicator(fx['dt'], fx['fx'], name="分型", row=1, line_width=1.2, visible=True)
            kline.add_scatter_indicator(bi['dt'], bi['bi'], name="笔", row=1, line_width=1.5)

        kline.add_sma(df, ma_seq=(5, 20, 120, 240), row=1, visible=False, line_width=1)
        kline.add_vol(df, row=2, line_width=1)
        kline.add_macd(df, row=3, line_width=1)

        for pos in trader.positions:
            bs_df = pd.DataFrame([x for x in pos.operates if x['dt'] >= c.bars_raw[0].dt])
            if not bs_df.empty:
                bs_df['dt'] = bs_df['dt'].apply(lambda x: freq_end_time(x, Freq(freq)))
                bs_df['tag'] = bs_df['op'].apply(lambda x: 'triangle-up' if x == Operate.LO else 'triangle-down')
                bs_df['color'] = bs_df['op'].apply(lambda x: 'red' if x == Operate.LO else 'silver')
                kline.add_scatter_indicator(
                    bs_df['dt'],
                    bs_df['price'],
                    name=pos.name,
                    text=bs_df['op_desc'],
                    row=1,
                    mode='text+markers',
                    marker_size=15,
                    marker_symbol=bs_df['tag'],
                    marker_color=bs_df['color'],
                )

        with tabs[i]:
            config = {
                "scrollZoom": True,
                "displayModeBar": True,
                "displaylogo": False,
                'modeBarButtonsToRemove': [
                    'toggleSpikelines',
                    'select2d',
                    'zoomIn2d',
                    'zoomOut2d',
                    'lasso2d',
                    'autoScale2d',
                    'hoverClosestCartesian',
                    'hoverCompareCartesian',
                ],
            }
            st.plotly_chart(kline.fig, use_container_width=True, config=config)
        i += 1

    # 信号页
    with tabs[i]:
        if len(trader.s):
            s = {k: v for k, v in trader.s.items() if len(k.split('_')) == 3}
            st.write(s)
    i += 1

    with tabs[i]:
        df = pd.DataFrame([x.evaluate() for x in trader.positions])
        st.dataframe(df, use_container_width=True)

        with st.expander("分别查看多头和空头的表现", expanded=False):
            df1 = pd.DataFrame([x.evaluate('多头') for x in trader.positions])
            st.dataframe(df1, use_container_width=True)

            df2 = pd.DataFrame([x.evaluate('空头') for x in trader.positions])
            st.dataframe(df2, use_container_width=True)

    i += 1
    with tabs[i]:
        for file in files:
            st.subheader(file.name)
            st.write(json.loads(file.getvalue().decode("utf-8")))

if files:
    if st.session_state.run:
        trader = CzscTrader(bg=bg, positions=deepcopy(tactic.positions), signals_config=deepcopy(tactic.signals_config))
        bars1 = bars_right[0 : st.session_state.bar_edt_index].copy()
        while bars1:
            bar_ = bars1.pop(0)
            trader.on_bar(bar_)

        bars2 = bars_right[st.session_state.bar_edt_index :].copy()
        with st.empty():
            while bars2:
                bar_ = bars2.pop(0)
                trader.on_bar(bar_)
                show_trader(trader)
                st.session_state.bar_edt_index += 1

    else:
        trader = CzscTrader(bg=bg, positions=deepcopy(tactic.positions), signals_config=deepcopy(tactic.signals_config))
        bars2 = bars_right[0 : st.session_state.bar_edt_index].copy()
        with st.empty():
            while bars2:
                bar_ = bars2.pop(0)
                trader.on_bar(bar_)
            show_trader(trader)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/5/6 13:24
describe: 提供一些策略的编写案例

以 trader_ 开头的是择时交易策略案例
"""
import os
import time
import shutil
import hashlib
import pandas as pd
from tqdm import tqdm
from copy import deepcopy
from datetime import timedelta
from abc import ABC, abstractmethod
from loguru import logger
from czsc.objects import RawBar, List, Operate, Signal, Factor, Event, Position
from czsc.traders.base import CzscTrader
from czsc.traders.sig_parse import get_signals_freqs, get_signals_config
from czsc.utils import x_round, freqs_sorted, BarGenerator, dill_dump, save_json, read_json


class CzscStrategyBase(ABC):
    """
    择时交易策略的要素：

    1. 交易品种以及该品种对应的参数
    2. K线周期列表
    3. 交易信号参数配置
    4. 持仓策略列表
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.signals_module_name = kwargs.get("signals_module_name", "czsc.signals")

    @property
    def symbol(self):
        """交易标的"""
        return self.kwargs["symbol"]

    @property
    def unique_signals(self):
        """所有持仓策略中的交易信号列表"""
        sig_seq = []
        for pos in self.positions:  # type: ignore
            sig_seq.extend(pos.unique_signals)
        return list(set(sig_seq))

    @property
    def signals_config(self):
        """交易信号参数配置"""
        return get_signals_config(self.unique_signals, self.signals_module_name)

    @property
    def freqs(self):
        """K线周期列表"""
        return get_signals_freqs(self.unique_signals)

    @property
    def sorted_freqs(self):
        """排好序的 K 线周期列表"""
        return freqs_sorted(self.freqs)

    @property
    def base_freq(self):
        """基础 K 线周期"""
        return self.sorted_freqs[0]

    @abstractmethod
    def positions(self) -> List[Position]:
        """持仓策略列表"""
        raise NotImplementedError

    def init_bar_generator(self, bars: List[RawBar], **kwargs):
        """使用策略定义初始化一个 BarGenerator 对象

        :param bars: 基础周期K线
        :param kwargs:
            bg   已经初始化好的BarGenerator对象，如果传入了bg，则忽略sdt和n参数
            sdt  初始化开始日期
            n    初始化最小K线数量
        :return:
        """
        base_freq = str(bars[0].freq.value)
        bg: BarGenerator = kwargs.get("bg", None)
        if base_freq in self.sorted_freqs:
            freqs = self.sorted_freqs[1:]
        else:
            freqs = self.sorted_freqs

        if bg is None:
            sdt = pd.to_datetime(kwargs.get("sdt", "20200101"))
            n = int(kwargs.get("n", 500))
            bg = BarGenerator(base_freq, freqs=freqs)

            # 拆分基础周期K线，sdt 之前的用来初始化BarGenerator，随后的K线是 trader 初始化区间
            bars_init = [x for x in bars if x.dt <= sdt]
            if len(bars_init) > n:
                bars1 = bars_init
                bars2 = [x for x in bars if x.dt > sdt]
            else:
                bars1 = bars[:n]
                bars2 = bars[n:]

            for bar in bars1:
                bg.update(bar)

            return bg, bars2
        else:
            assert bg.base_freq == bars[-1].freq.value, "BarGenerator 的基础周期和 bars 的基础周期不一致"
            bars2 = [x for x in bars if x.dt > bg.end_dt]
            return bg, bars2

    def init_trader(self, bars: List[RawBar], **kwargs) -> CzscTrader:
        """使用策略定义初始化一个 CzscTrader 对象

        **注意：** 这里会将所有持仓策略在 sdt 之后的交易信号计算出来并缓存在持仓策略实例内部，所以初始化的过程本身也是回测的过程。

        :param bars: 基础周期K线
        :param kwargs:
            bg   已经初始化好的BarGenerator对象，如果传入了bg，则忽略sdt和n参数
            sdt  初始化开始日期
            n    初始化最小K线数量
        :return: 完成策略初始化后的 CzscTrader 对象
        """
        bg, bars2 = self.init_bar_generator(bars, **kwargs)
        trader = CzscTrader(
            bg=bg, positions=deepcopy(self.positions), signals_config=deepcopy(self.signals_config), **kwargs
        )
        for bar in bars2:
            trader.on_bar(bar)
        return trader

    def backtest(self, bars: List[RawBar], **kwargs) -> CzscTrader:
        trader = self.init_trader(bars, **kwargs)
        return trader

    def dummy(self, sigs: List[dict], **kwargs) -> CzscTrader:
        """使用信号缓存进行策略回测

        :param sigs: 信号缓存，一般指 generate_czsc_signals 函数计算的结果缓存
        :return: 完成策略回测后的 CzscTrader 对象
        """
        sleep_time = kwargs.get("sleep_time", 0)
        sleep_step = kwargs.get("sleep_step", 1000)

        trader = CzscTrader(positions=deepcopy(self.positions))
        for i, sig in tqdm(enumerate(sigs), desc=f"回测 {self.symbol} {self.sorted_freqs}"):
            trader.on_sig(sig)

            if i % sleep_step == 0:
                time.sleep(sleep_time)

        return trader

    def replay(self, bars: List[RawBar], res_path, **kwargs):
        """交易策略交易过程回放

        :param bars: 基础周期K线
        :param res_path: 结果目录
        :param kwargs:
            bg   已经初始化好的BarGenerator对象，如果传入了bg，则忽略sdt和n参数
            sdt  初始化开始日期
            n    初始化最小K线数量
        :return:
        """
        if kwargs.get("refresh", False):
            shutil.rmtree(res_path, ignore_errors=True)

        exist_ok = kwargs.get("exist_ok", False)
        if os.path.exists(res_path) and not exist_ok:
            logger.warning(f"结果文件夹存在且不允许覆盖：{res_path}，如需执行，请先删除文件夹")
            return
        os.makedirs(res_path, exist_ok=exist_ok)

        bg, bars2 = self.init_bar_generator(bars, **kwargs)
        trader = CzscTrader(
            bg=bg, positions=deepcopy(self.positions), signals_config=deepcopy(self.signals_config), **kwargs
        )
        for position in trader.positions:
            pos_path = os.path.join(res_path, position.name)
            os.makedirs(pos_path, exist_ok=exist_ok)

        for bar in bars2:
            trader.on_bar(bar)
            for position in trader.positions:
                pos_path = os.path.join(res_path, position.name)

                if position.operates and position.operates[-1]["dt"] == bar.dt:
                    op = position.operates[-1]
                    _dt = op["dt"].strftime("%Y%m%d#%H%M")
                    file_name = f"{_dt}_{op['op'].value}_{op['bid']}_{x_round(op['price'], 2)}_{op['op_desc']}.html"
                    file_html = os.path.join(pos_path, file_name)
                    trader.take_snapshot(file_html)
                    logger.info(f"{file_html}")

        for position in trader.positions:
            logger.info(
                f"{position.name}  "
                f"\n 多空合并：{position.evaluate()} "
                f"\n 多头表现：{position.evaluate('多头')} "
                f"\n 空头表现：{position.evaluate('空头')}"
            )

        file_trader = os.path.join(res_path, "trader.ct")
        try:
            dill_dump(trader, file_trader)
            logger.info(f"交易对象保存到：{file_trader}")
        except Exception as e:
            logger.error(f"交易对象保存失败：{e}；通常的原因是交易对象中包含了不支持序列化的对象，比如函数")
        return trader

    def check(self, bars: List[RawBar], res_path, **kwargs):
        """检查交易策略中的信号是否正确

        :param bars: 基础周期K线
        :param res_path: 结果目录
        :param kwargs:
            bg   已经初始化好的BarGenerator对象，如果传入了bg，则忽略sdt和n参数
            sdt  初始化开始日期
            n    初始化最小K线数量
        :return:
        """
        if kwargs.get("refresh", False):
            shutil.rmtree(res_path, ignore_errors=True)

        exist_ok = kwargs.get("exist_ok", False)
        if os.path.exists(res_path) and not exist_ok:
            logger.warning(f"结果文件夹存在且不允许覆盖：{res_path}，如需执行，请先删除文件夹")
            return
        os.makedirs(res_path, exist_ok=exist_ok)

        # 第一遍执行，获取信号
        bg, bars2 = self.init_bar_generator(bars, **kwargs)
        trader = CzscTrader(
            bg=bg, positions=deepcopy(self.positions), signals_config=deepcopy(self.signals_config), **kwargs
        )

        _signals = []
        for bar in bars2:
            trader.on_bar(bar)
            _signals.append(trader.s)

        for position in trader.positions:
            print(f"{position.name}: {position.evaluate()}")

        df = pd.DataFrame(_signals)
        df.to_excel(os.path.join(res_path, "signals.xlsx"), index=False)
        unique_signals = {}
        for col in [x for x in df.columns if len(x.split("_")) == 3]:
            unique_signals[col] = [Signal(f"{col}_{v}") for v in df[col].unique() if "其他" not in v]

        print("\n", "+" * 100)
        for key, values in unique_signals.items():
            print(f"\n{key}:")
            for value in values:
                print(f"- {value}")
        print("\n", "+" * 100)

        # 第二遍执行，检查信号，生成html
        bg, bars2 = self.init_bar_generator(bars, **kwargs)
        trader = CzscTrader(
            bg=bg, positions=deepcopy(self.positions), signals_config=deepcopy(self.signals_config), **kwargs
        )

        # 记录每个信号最后一次出现的时间
        last_sig_dt = {y.key: trader.end_dt for x in unique_signals.values() for y in x}
        delta_days = kwargs.get("delta_days", 1)

        for bar in bars2:
            trader.on_bar(bar)

            for key, values in unique_signals.items():
                html_path = os.path.join(res_path, key)
                os.makedirs(html_path, exist_ok=True)

                for signal in values:
                    if bar.dt - last_sig_dt[signal.key] > timedelta(days=delta_days) and signal.is_match(trader.s):
                        file_html = f"{bar.dt.strftime('%Y%m%d_%H%M')}_{signal.signal}.html"
                        file_html = os.path.join(html_path, file_html)
                        print(file_html)
                        trader.take_snapshot(file_html, height=kwargs.get("height", "680px"))
                        last_sig_dt[signal.key] = bar.dt

    def save_positions(self, path):
        """保存持仓策略配置

        :param path: 结果路径
        :return: None
        """
        os.makedirs(path, exist_ok=True)
        for pos in self.positions:
            pos_ = pos.dump()
            pos_.pop("symbol")
            hash_code = hashlib.md5(str(pos_).encode()).hexdigest()
            pos_["md5"] = hash_code
            save_json(pos_, os.path.join(path, f"{pos_['name']}.json"))

    def load_positions(self, files: List, check=True) -> List[Position]:
        """从配置文件中加载持仓策略

        :param files: 以json格式保存的持仓策略文件列表
        :param check: 是否校验 MD5 值，默认为 True
        :return: 持仓策略列表
        """
        positions = []
        for file in files:
            pos = read_json(file)
            md5 = pos.pop("md5")
            if check:
                assert md5 == hashlib.md5(str(pos).encode()).hexdigest()
            pos["symbol"] = self.symbol
            positions.append(Position.load(pos))
        return positions


class CzscJsonStrategy(CzscStrategyBase):
    """仅传入Json配置的Positions就完成策略创建

    必须参数：
        files_position: 以 json 文件配置的策略，每个json文件对应一个持仓策略配置
        check_position: 是否对 json 持仓策略进行 MD5 校验，默认为 True
    """

    @property
    def positions(self):
        files = self.kwargs["files_position"]
        check = self.kwargs.get("check_position", True)
        return self.load_positions(files, check)


class CzscStrategyExample2(CzscStrategyBase):
    """仅传入Positions就完成策略创建"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def positions(self):
        return [self.create_pos_a(), Position.load(self.create_pos_b())]

    def create_pos_a(self):
        opens = [
            Event(
                name="开多",
                operate=Operate.LO,
                factors=[
                    Factor(
                        name="15分钟向下笔停顿",
                        signals_all=[
                            Signal("15分钟_D0停顿分型_BE辅助V230106_看多_强_任意_0"),
                        ],
                    )
                ],
            ),
            Event(
                name="开空",
                operate=Operate.SO,
                factors=[
                    Factor(
                        name="15分钟向上笔停顿",
                        signals_all=[
                            Signal("15分钟_D0停顿分型_BE辅助V230106_看空_强_任意_0"),
                        ],
                    )
                ],
            ),
        ]
        pos = Position(
            name="15分钟笔停顿", symbol=self.symbol, opens=opens, exits=None, interval=0, timeout=20, stop_loss=100, T0=True
        )
        return pos

    def create_pos_b(self):
        """从 json文件 / dict 中加载 Position"""
        return {
            "symbol": self.symbol,
            "name": "15分钟笔停顿B",
            "opens": [
                {
                    "name": "开多",
                    "operate": "开多",
                    "signals_all": [],
                    "signals_any": [],
                    "signals_not": [],
                    "factors": [
                        {
                            "name": "15分钟向下笔停顿",
                            "signals_all": ["15分钟_D0停顿分型_BE辅助V230106_看多_强_任意_0"],
                            "signals_any": [],
                            "signals_not": [],
                        }
                    ],
                },
                {
                    "name": "开空",
                    "operate": "开空",
                    "signals_all": [],
                    "signals_any": [],
                    "signals_not": [],
                    "factors": [
                        {
                            "name": "15分钟向上笔停顿",
                            "signals_all": ["15分钟_D0停顿分型_BE辅助V230106_看空_强_任意_0"],
                            "signals_any": [],
                            "signals_not": [],
                        }
                    ],
                },
            ],
            "exits": [],
            "interval": 0,
            "timeout": 20,
            "stop_loss": 100,
            "T0": True,
        }


def create_single_ma_long(symbol, ma_name, is_stocks=False, **kwargs) -> Position:
    """单均线多头策略

    https://czsc.readthedocs.io/en/latest/api/czsc.signals.tas_ma_base_V230313.html

    :param symbol:
    :param ma_name:
    :param is_stocks:
    :param kwargs:
    :return:
    """
    mo = int(kwargs.get("max_overlap", 5))
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)
    
    opens = [
        {
            "operate": "开多",
            "signals_not": [],
            "signals_all": [],
            "factors": [{"name": f"{ma_name}多头", "signals_all": [f"{freq}_D1#{ma_name}MO{mo}_BS辅助V230313_看多_任意_任意_0"]}],
        },
    ]

    exits = [
        {
            "operate": "平多",
            "signals_all": [],
            "signals_not": [],
            "factors": [{"name": f"{ma_name}空头", "signals_all": [f"{freq}_D1#{ma_name}MO{mo}_BS辅助V230313_看空_任意_任意_0"]}],
        },
    ]
    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [
            f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0",
            f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0",
        ]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}{ma_name}多头基准"
    else:
        # 非A股：都行。加入这个条件，主要是为了约束策略使用基础周期K线
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}{ma_name}多头基准"

    T0 = kwargs.get("T0", False)
    pos_name = f"{pos_name}T0" if T0 else f"{pos_name}"

    pos = Position(
        name=pos_name,
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=kwargs.get("interval", 3600 * 2),
        timeout=kwargs.get("timeout", 16 * 30),
        stop_loss=kwargs.get("stop_loss", 300),
        T0=T0,
    )
    return pos


def create_single_ma_short(symbol, ma_name, is_stocks=False, **kwargs) -> Position:
    """单均线空头策略

    https://czsc.readthedocs.io/en/latest/api/czsc.signals.tas_ma_base_V230313.html

    :param symbol: 标的代码
    :param ma_name: 均线名称，如 SMA#5
    :param is_stocks: 是否是 A 股
    :param kwargs: 其他参数
    :return:
    """
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)
    mo = int(kwargs.get("max_overlap", 5))

    opens = [
        {
            "operate": "开空",
            "signals_not": [],
            "signals_all": [],
            "factors": [{"name": f"{ma_name}空头", "signals_all": [f"{freq}_D1#{ma_name}MO{mo}_BS辅助V230313_看空_任意_任意_0"]}],
        },
    ]

    exits = [
        {
            "operate": "平空",
            "signals_all": [],
            "signals_not": [],
            "factors": [{"name": f"{ma_name}多头", "signals_all": [f"{freq}_D1#{ma_name}MO{mo}_BS辅助V230313_看多_任意_任意_0"]}],
        },
    ]
    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [
            f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0",
            f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0",
        ]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}{ma_name}空头基准"
    else:
        # 非A股：都行
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}{ma_name}空头基准"

    T0 = kwargs.get("T0", False)
    pos_name = f"{pos_name}T0" if T0 else f"{pos_name}"

    pos = Position(
        name=pos_name,
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=kwargs.get("interval", 3600 * 2),
        timeout=kwargs.get("timeout", 16 * 30),
        stop_loss=kwargs.get("stop_loss", 300),
        T0=T0,
    )
    return pos


def create_macd_short(symbol, is_stocks=False, **kwargs) -> Position:
    """MACD 空头策略

    https://czsc.readthedocs.io/en/latest/api/czsc.signals.tas_macd_base_V230320.html

    :param symbol: 标的代码
    :param is_stocks: 是否是 A 股
    :param kwargs: 其他参数
    :return:
    """
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)

    mo = int(kwargs.get("max_overlap", 5))
    opens = [
        {
            "operate": "开空",
            "signals_all": [],
            "signals_not": [],
            "factors": [
                {"name": f"MACD空头", "signals_all": [f"{freq}_D1MACD12#26#9MO{mo}#MACD_BS辅助V230320_空头_任意_任意_0"]}
            ],
        },
    ]

    exits = [
        {
            "operate": "平空",
            "signals_all": [],
            "signals_not": [],
            "factors": [{"name": "MACD多头", "signals_all": [f"{freq}_D1MACD12#26#9MO{mo}#MACD_BS辅助V230320_多头_任意_任意_0"]}],
        },
    ]
    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [
            f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0",
            f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0",
        ]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}MACD空头基准"
    else:
        # 非A股：都行
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}MACD空头基准"

    T0 = kwargs.get("T0", False)
    pos_name = f"{pos_name}T0" if T0 else f"{pos_name}"

    pos = Position(
        name=pos_name,
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=kwargs.get("interval", 3600 * 2),
        timeout=kwargs.get("timeout", 16 * 30),
        stop_loss=kwargs.get("stop_loss", 300),
        T0=T0,
    )
    return pos


def create_macd_long(symbol, is_stocks=False, **kwargs) -> Position:
    """MACD 多头策略

    https://czsc.readthedocs.io/en/latest/api/czsc.signals.tas_macd_base_V230320.html

    :param symbol: 标的代码
    :param is_stocks: 是否是 A 股
    :param kwargs: 其他参数
        - base_freq: 基础级别
        - freq: 信号级别
        - max_overlap: 最大重叠数
        - T0: 是否是 T0 策略
    :return:
    """
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)
    mo = int(kwargs.get("max_overlap", 5))
    T0 = kwargs.get("T0", False)

    opens = [
        {
            "operate": "开多",
            "signals_all": [],
            "signals_not": [],
            "factors": [{"name": "MACD多头", "signals_all": [f"{freq}_D1MACD12#26#9MO{mo}#MACD_BS辅助V230320_多头_任意_任意_0"]}],
        },
    ]

    exits = [
        {
            "operate": "平多",
            "signals_all": [],
            "signals_not": [],
            "factors": [
                {"name": f"MACD空头", "signals_all": [f"{freq}_D1MACD12#26#9MO{mo}#MACD_BS辅助V230320_空头_任意_任意_0"]}
            ],
        },
    ]
    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [
            f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0",
            f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0",
        ]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}MACD多头基准"
    else:
        # 非A股：都行
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}MACD多头基准"

    pos = Position(
        name=f"{pos_name}T0" if T0 else f"{pos_name}",
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=kwargs.get("interval", 3600 * 2),
        timeout=kwargs.get("timeout", 16 * 30),
        stop_loss=kwargs.get("stop_loss", 300),
        T0=T0,
    )
    return pos


def create_cci_long(symbol, is_stocks=False, **kwargs) -> Position:
    """CCI基础多头策略

    用到的信号函数列表：

    1. https://czsc.readthedocs.io/en/latest/api/czsc.signals.tas_cci_base_V230402.html
    2. https://czsc.readthedocs.io/en/latest/api/czsc.signals.bar_zdt_V230331.html

    :param symbol: 标的代码
    :param is_stocks: 是否是 A 股
    :param kwargs: 其他参数
        - base_freq: 基础级别
        - freq: 信号级别
        - cci_timeperiod: CCI 周期
        - T0: 是否是 T0 策略
        - interval: 同向开仓间隔时间
        - timeout: 超时出场时间
    :return:
    """
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)
    cci_timeperiod = kwargs.get("cci_timeperiod", 14)
    T0 = kwargs.get("T0", False)
    interval = kwargs.get("interval", 3600 * 2)
    timeout = kwargs.get("timeout", 16 * 30)
    stop_loss = kwargs.get("stop_loss", 300)

    opens = [
        {
            "operate": "开多",
            "signals_all": [],
            "signals_any": [],
            "signals_not": [],
            "factors": [
                {
                    "name": "CCI看多",
                    "signals_all": [f"{freq}_D1CCI{cci_timeperiod}#3#10_BS辅助V230402_多头_任意_任意_0"],
                    "signals_any": [],
                    "signals_not": [],
                },
            ],
        }
    ]

    exits = [
        {
            "operate": "平多",
            "signals_all": [],
            "signals_any": [],
            "signals_not": [],
            "factors": [
                {
                    "name": "CCI看空",
                    "signals_all": [f"{freq}_D1CCI{cci_timeperiod}#3#60_BS辅助V230402_空头_任意_任意_0"],
                    "signals_any": [],
                    "signals_not": [],
                },
            ],
        }
    ]

    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [
            f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0",
            f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0",
        ]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}CCI多头基准"
    else:
        # 非A股：都行
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}CCI多头基准"

    pos = Position(
        name=f"{pos_name}T0" if T0 else f"{pos_name}",
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=interval,
        timeout=timeout,
        stop_loss=stop_loss,
        T0=T0,
    )
    return pos


def create_cci_short(symbol, is_stocks=False, **kwargs) -> Position:
    """CCI基础空头策略

    用到的信号函数列表：

    1. https://czsc.readthedocs.io/en/latest/api/czsc.signals.tas_cci_base_V230402.html
    2. https://czsc.readthedocs.io/en/latest/api/czsc.signals.bar_zdt_V230331.html

    :param symbol: 标的代码
    :param is_stocks: 是否是 A 股
    :param kwargs: 其他参数
        - base_freq: 基础级别
        - freq: 信号级别
        - cci_timeperiod: CCI 周期
        - T0: 是否是 T0 策略
        - interval: 同向开仓间隔时间
        - timeout: 超时出场时间
    :return:
    """
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)
    cci_timeperiod = kwargs.get("cci_timeperiod", 14)
    T0 = kwargs.get("T0", False)
    interval = kwargs.get("interval", 3600 * 2)
    timeout = kwargs.get("timeout", 16 * 30)
    stop_loss = kwargs.get("stop_loss", 300)

    opens = [
        {
            "operate": "开空",
            "signals_all": [],
            "signals_any": [],
            "signals_not": [],
            "factors": [
                {
                    "name": "CCI看空",
                    "signals_all": [f"{freq}_D1CCI{cci_timeperiod}#3#10_BS辅助V230402_空头_任意_任意_0"],
                    "signals_any": [],
                    "signals_not": [],
                },
            ],
        }
    ]

    exits = [
        {
            "operate": "平空",
            "signals_all": [],
            "signals_any": [],
            "signals_not": [],
            "factors": [
                {
                    "name": "CCI看多",
                    "signals_all": [f"{freq}_D1CCI{cci_timeperiod}#3#60_BS辅助V230402_多头_任意_任意_0"],
                    "signals_any": [],
                    "signals_not": [],
                },
            ],
        }
    ]

    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [
            f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0",
            f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0",
        ]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}CCI空头基准"
    else:
        # 非A股：都行
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}CCI空头基准"

    pos = Position(
        name=f"{pos_name}T0" if T0 else f"{pos_name}",
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=interval,
        timeout=timeout,
        stop_loss=stop_loss,
        T0=T0,
    )
    return pos


def create_emv_long(symbol, is_stocks=False, **kwargs) -> Position:
    """EMV 多头策略

    :param symbol: 标的代码
    :param is_stocks: 是否是 A 股
    :param kwargs: 其他参数

        - base_freq: 基础级别
        - freq: 信号级别
        - T0: 是否是 T0 策略

    :return:
    """
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)
    di = int(kwargs.get("di", 1))
    T0 = kwargs.get("T0", False)
    timeout = int(kwargs.get("timeout", 100))
    stop_loss = int(kwargs.get("stop_loss", 300))
    interval = int(kwargs.get("interval", 3600 * 2))    # 同向开仓时间间隔，单位：秒；默认 2 小时，一般不用修改

    opens = [
        {
            "operate": "开多",
            "signals_not": [],
            "factors": [{"name": "EMV多头", "signals_all": [f"{freq}_D{di}_EMV简易波动V230605_看多_任意_任意_0"]}],
        }
    ]

    exits = [
        {
            "operate": "平多",
            "signals_not": [],
            "factors": [{"name": f"EMV空头", "signals_all": [f"{freq}_D{di}_EMV简易波动V230605_看空_任意_任意_0"]}],
        }
    ]
    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0", f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0"]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}EMV多头基准"
    else:
        # 非A股：都行
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}EMV多头基准"

    pos = Position(
        name=f"{pos_name}T0" if T0 else f"{pos_name}",
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=interval,
        timeout=timeout,
        stop_loss=stop_loss,
        T0=T0,
    )
    return pos


def create_emv_short(symbol, is_stocks=False, **kwargs) -> Position:
    """EMV 空头策略

    :param symbol: 标的代码
    :param is_stocks: 是否是 A 股
    :param kwargs: 其他参数

        - base_freq: 基础级别
        - freq: 信号级别
        - T0: 是否是 T0 策略

    :return:
    """
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)
    di = int(kwargs.get("di", 1))
    T0 = kwargs.get("T0", False)
    timeout = int(kwargs.get("timeout", 100))
    stop_loss = int(kwargs.get("stop_loss", 300))
    interval = int(kwargs.get("interval", 3600 * 2))    # 同向开仓时间间隔，单位：秒；默认 2 小时，一般不用修改

    opens = [
                {
            "operate": "开空",
            "signals_not": [],
            "factors": [{"name": f"EMV空头", "signals_all": [f"{freq}_D{di}_EMV简易波动V230605_看空_任意_任意_0"]}],
        }

    ]

    exits = [
        {
            "operate": "平空",
            "signals_not": [],
            "factors": [{"name": "EMV多头", "signals_all": [f"{freq}_D{di}_EMV简易波动V230605_看多_任意_任意_0"]}],
        }
    ]
    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0", f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0"]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}EMV空头基准"
    else:
        # 非A股：都行
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}EMV空头基准"

    pos = Position(
        name=f"{pos_name}T0" if T0 else f"{pos_name}",
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=interval,
        timeout=timeout,
        stop_loss=stop_loss,
        T0=T0,
    )
    return pos


def create_third_buy_long(symbol, is_stocks=False, **kwargs) -> Position:
    """缠中说禅三买多头策略

    https://czsc.readthedocs.io/en/latest/api/czsc.signals.cxt_five_bi_V230619.html
    https://czsc.readthedocs.io/en/latest/api/czsc.signals.cxt_three_bi_V230618.html

    :param symbol: 标的代码
    :param is_stocks: 是否是 A 股
    :param kwargs: 其他参数

        - base_freq: 基础级别
        - freq: 信号级别
        - T0: 是否是 T0 策略

    :return:
    """
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)
    T0 = kwargs.get("T0", False)
    timeout = int(kwargs.get("timeout", 100))
    stop_loss = int(kwargs.get("stop_loss", 300))
    interval = int(kwargs.get("interval", 3600 * 2))  # 同向开仓时间间隔，单位：秒；默认 2 小时，一般不用修改

    opens = [
        {
            "operate": "开多",
            "signals_all": [],
            "signals_not": [],
            "factors": [{"name": "五笔三买", "signals_all": [f"{freq}_D1五笔_形态V230619_类三买_任意_任意_0"], "signals_any": []}],
        }
    ]

    exits = [
        {
            "operate": "平多",
            "signals_all": [],
            "signals_not": [],
            "factors": [
                {"name": "三笔向上盘背", "signals_all": [f"{freq}_D1三笔_形态V230618_向上盘背_任意_任意_0"], "signals_any": []},
                {"name": "三笔向上收敛", "signals_all": [f"{freq}_D1三笔_形态V230618_向上收敛_任意_任意_0"], "signals_any": []},
                {"name": "三笔向上扩张", "signals_all": [f"{freq}_D1三笔_形态V230618_向上扩张_任意_任意_0"], "signals_any": []},
                {"name": "三笔不创新高", "signals_all": [f"{freq}_D1三笔_形态V230618_向上奔走型_任意_任意_0"], "signals_any": []},
            ],
        }
    ]
    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0", f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0"]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}缠论三买多头基准"
    else:
        # 非A股：都行
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}缠论三买多头基准"

    pos = Position(
        name=f"{pos_name}T0" if T0 else f"{pos_name}",
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=interval,
        timeout=timeout,
        stop_loss=stop_loss,
        T0=T0,
    )
    return pos


def create_third_sell_short(symbol, is_stocks=False, **kwargs) -> Position:
    """缠中说禅三卖空头策略

    :param symbol: 标的代码
    :param is_stocks: 是否是 A 股
    :param kwargs: 其他参数

        - base_freq: 基础级别
        - freq: 信号级别
        - T0: 是否是 T0 策略

    :return:
    """
    freq = kwargs.get("freq", "15分钟")
    base_freq = kwargs.get("base_freq", freq)
    T0 = kwargs.get("T0", False)
    timeout = int(kwargs.get("timeout", 100))
    stop_loss = int(kwargs.get("stop_loss", 300))
    interval = int(kwargs.get("interval", 3600 * 2))  # 同向开仓时间间隔，单位：秒；默认 2 小时，一般不用修改

    opens = [
        {
            "operate": "开空",
            "signals_all": [],
            "signals_not": [],
            "factors": [{"name": "五笔三卖", "signals_all": [f"{freq}_D1五笔_形态V230619_类三卖_任意_任意_0"], "signals_any": []}],
        }
    ]

    exits = [
        {
            "operate": "平空",
            "signals_all": [],
            "signals_not": [],
            "factors": [
                {"name": "三笔向下盘背", "signals_all": [f"{freq}_D1三笔_形态V230618_向下盘背_任意_任意_0"], "signals_any": []},
                {"name": "三笔向下收敛", "signals_all": [f"{freq}_D1三笔_形态V230618_向下收敛_任意_任意_0"], "signals_any": []},
                {"name": "三笔向下扩张", "signals_all": [f"{freq}_D1三笔_形态V230618_向下扩张_任意_任意_0"], "signals_any": []},
                {"name": "三笔不创新低", "signals_all": [f"{freq}_D1三笔_形态V230618_向下奔走型_任意_任意_0"], "signals_any": []},
            ],
        }
    ]
    if is_stocks:
        # A股：涨跌停不交易
        zdt_sigs = [f"{base_freq}_D1_涨跌停V230331_跌停_任意_任意_0", f"{base_freq}_D1_涨跌停V230331_涨停_任意_任意_0"]
        opens[0]["signals_not"].extend(zdt_sigs)
        exits[0]["signals_not"].extend(zdt_sigs)
        pos_name = f"A股{freq}缠论三卖空头基准"
    else:
        # 非A股：都行
        opens[0]["signals_all"].append(f"{base_freq}_D1_涨跌停V230331_任意_任意_任意_0")
        pos_name = f"{freq}缠论三卖空头基准"

    pos = Position(
        name=f"{pos_name}T0" if T0 else f"{pos_name}",
        symbol=symbol,
        opens=[Event.load(x) for x in opens],
        exits=[Event.load(x) for x in exits],
        interval=interval,
        timeout=timeout,
        stop_loss=stop_loss,
        T0=T0,
    )
    return pos



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/5/11 20:49
describe: CZSC策略回放
"""
import os
os.environ['czsc_max_bi_num'] = '20'
os.environ['czsc_research_cache'] = r"D:\CZSC投研数据"
os.environ['signals_module_name'] = 'czsc.signals'
import inspect
import streamlit as st
import pandas as pd
from typing import List
from copy import deepcopy
from czsc.utils.bar_generator import freq_end_time
from czsc.connectors.research import get_symbols, get_raw_bars
from czsc import CzscStrategyBase, Position, CzscTrader, KlineChart, Freq, Operate

st.set_page_config(layout="wide")

# ======================================================================================================================
# 将策略的代码放在这里，命名是 Strategy
# ----------------------------------------------------------------------------------------------------------------------

# 可以直接导入策略类，也可以直接写在这里
from czsc.strategies import CzscStrategyExample2 as Strategy


# class Strategy(CzscStrategyBase):
#
#     @staticmethod
#     def create_long_bi(symbol, freq1='5分钟', freq2='60分钟'):
#         """
#
#         :param symbol:
#         :param freq1:
#         :param freq2:
#         :return:
#         """
#         _pos_name = f'{freq1}#{freq2}向下笔多头'
#         _pos = {'symbol': symbol,
#                 'name': _pos_name,
#                 'opens': [
#                     {'operate': '开多',
#                      'factors': [
#                          {'name': f'{freq1}向下笔停顿分型',
#                           'signals_all': [
#                               f'{freq1}_D0停顿分型_BE辅助V230106_看多_任意_任意_0',
#                               f'{freq2}_D0停顿分型_BE辅助V230106_看多_任意_任意_0'
#                           ]},
#
#                          {'name': f'{freq1}向下笔验证分型',
#                           'signals_all': [
#                               f'{freq1}_D0验证分型_BE辅助V230107_看多_任意_任意_0'
#                           ]},
#                      ]}
#                 ],
#                 'exits': [
#                     {'operate': '平多',
#                      'factors': [
#                          {'name': f'{freq1}停顿分型止盈',
#                           'signals_all': [
#                               f'{freq1}_D0停顿分型_BE辅助V230106_看空_任意_任意_0'
#                           ]},
#
#                          {'name': f'{freq1}验证分型止盈',
#                           'signals_all': [
#                               f'{freq1}_D0验证分型_BE辅助V230107_看空_任意_任意_0'
#                           ]},
#                      ]},
#                 ],
#                 'interval': 7200,
#                 'timeout': 1000,
#                 'stop_loss': 300,
#                 'T0': True}
#
#         return Position.load(_pos)
#
#     @property
#     def positions(self) -> List[Position]:
#         _pos_list = [
#             self.create_long_bi(symbol=self.symbol, freq1='5分钟', freq2='60分钟'),
#             # self.create_long_bi(symbol=self.symbol, freq1='15分钟', freq2='60分钟'),
#         ]
#         return _pos_list


# ======================================================================================================================
# 以下代码不用修改
# ----------------------------------------------------------------------------------------------------------------------
if 'bar_edt_index' not in st.session_state:
    st.session_state.bar_edt_index = 1

if 'run' not in st.session_state:
    st.session_state.run = False

if 'date_change' not in st.session_state:
    st.session_state.date_change = True


def update_bar_edt():
    bar_edt_ = pd.Timestamp(st.session_state.bar_edt)
    i = 1
    for x in bars_right:
        if x.dt >= bar_edt_:
            break
        i += 1
    st.session_state.bar_edt_index = i


def update_date_change():
    st.session_state.date_change = True
    st.session_state.bar_edt_index = 1


@st.cache_data
def get_bars(symbol_, base_freq_, sdt_, edt_):
    delta_days_map = {
        '1分钟': 6,
        '5分钟': 45,
        '15分钟': 90,
        '30分钟': 180,
        '60分钟': 365,
        '日线': 365 * 3,
        '周线': 365 * 7,
    }
    return get_raw_bars(symbol_, base_freq_, sdt=sdt_ - pd.Timedelta(days=delta_days_map[base_freq_]), edt=edt_)


with st.sidebar:
    st.title("CZSC策略回放")
    symbol = st.selectbox("选择标的：", get_symbols('ALL'), index=0, on_change=update_date_change)
    sdt = st.date_input("开始日期：", value=pd.to_datetime('2022-01-01'), on_change=update_date_change)
    edt = st.date_input("结束日期：", value=pd.to_datetime('2022-12-31'), on_change=update_date_change)
    czsc_strategy = Strategy
    tactic: CzscStrategyBase = czsc_strategy(symbol=symbol, signals_module_name=os.environ['signals_module_name'])

    if st.session_state.date_change:
        bars = get_bars(symbol, tactic.base_freq, sdt, edt)
        bg, bars_right = tactic.init_bar_generator(bars, sdt=sdt)
        if len(bars_right) > 1000:
            bars_right = bars_right[:1000]  # 限制回放的数据量
            st.warning("回放数据量过大，已限制为1000根K线")
        bars_num = len(bars_right)
    else:
        st.session_state.date_change = False

    c1, c2, c3, c4 = st.columns([5, 5, 5, 5])
    with c1:
        button_run = st.button('播放')
        if button_run:
            st.session_state.run = True

    with c2:
        button_run = st.button('暂停')
        if button_run:
            st.session_state.run = False

    with c3:
        button_dec = st.button('左移')
        if button_dec:
            st.session_state.bar_edt_index -= 1
            st.session_state.bar_edt_index = max(st.session_state.bar_edt_index, 0)

    with c4:
        button_inc = st.button('右移')
        if button_inc:
            st.session_state.bar_edt_index += 1
            st.session_state.bar_edt_index = min(st.session_state.bar_edt_index, bars_num)

    bar_edt = bars_right[st.session_state.bar_edt_index - 1].dt
    st.text_input('定位到指定时间', value=bar_edt.strftime('%Y-%m-%d %H:%M'), key="bar_edt", on_change=update_bar_edt)

    st.success(f"共{bars_num}根K线, 时间范围：{bars_right[0].dt} - {bars_right[-1].dt}; 当前K线：{bar_edt}")


def show_trader(trader: CzscTrader):
    freqs = trader.freqs
    tabs = st.tabs(freqs + ['最后信号', '收益分析', '策略脚本'])

    i = 0
    for freq in freqs:
        c = trader.kas[freq]
        df = pd.DataFrame(c.bars_raw)
        kline = KlineChart(n_rows=3, title='', width="100%", height=800)
        kline.add_kline(df, name="")

        if len(c.bi_list) > 0:
            bi = pd.DataFrame([{'dt': x.fx_a.dt, "bi": x.fx_a.fx} for x in c.bi_list] +
                              [{'dt': c.bi_list[-1].fx_b.dt, "bi": c.bi_list[-1].fx_b.fx}])
            fx = pd.DataFrame([{'dt': x.dt, "fx": x.fx} for x in c.fx_list])
            kline.add_scatter_indicator(fx['dt'], fx['fx'], name="分型", row=1, line_width=1.2,
                                        visible=True)
            kline.add_scatter_indicator(bi['dt'], bi['bi'], name="笔", row=1, line_width=1.5)

        kline.add_sma(df, ma_seq=(5, 20, 120, 240), row=1, visible=False, line_width=1)
        kline.add_vol(df, row=2, line_width=1)
        kline.add_macd(df, row=3, line_width=1)

        for pos in trader.positions:
            bs_df = pd.DataFrame([x for x in pos.operates if x['dt'] >= c.bars_raw[0].dt])
            if not bs_df.empty:
                bs_df['dt'] = bs_df['dt'].apply(lambda x: freq_end_time(x, Freq(freq)))
                bs_df['tag'] = bs_df['op'].apply(lambda x: 'triangle-up' if x == Operate.LO else 'triangle-down')
                bs_df['color'] = bs_df['op'].apply(lambda x: 'red' if x == Operate.LO else 'silver')
                kline.add_scatter_indicator(bs_df['dt'], bs_df['price'], name=pos.name, text=bs_df['op_desc'], row=1,
                                            mode='text+markers', marker_size=15, marker_symbol=bs_df['tag'],
                                            marker_color=bs_df['color'])

        with tabs[i]:
            config = {
                "scrollZoom": True,
                "displayModeBar": True,
                "displaylogo": False,
                'modeBarButtonsToRemove': [
                    'toggleSpikelines',
                    'select2d',
                    'zoomIn2d',
                    'zoomOut2d',
                    'lasso2d',
                    'autoScale2d',
                    'hoverClosestCartesian',
                    'hoverCompareCartesian']}
            st.plotly_chart(kline.fig, use_container_width=True, config=config)
        i += 1

    # 信号页
    with tabs[i]:
        if len(trader.s):
            s = {k: v for k, v in trader.s.items() if len(k.split('_')) == 3}
            st.write(s)
    i += 1

    with tabs[i]:
        df = pd.DataFrame([x.evaluate() for x in trader.positions])
        st.dataframe(df, use_container_width=True)

        with st.expander("分别查看多头和空头的表现", expanded=False):
            df1 = pd.DataFrame([x.evaluate('多头') for x in trader.positions])
            st.dataframe(df1, use_container_width=True)

            df2 = pd.DataFrame([x.evaluate('空头') for x in trader.positions])
            st.dataframe(df2, use_container_width=True)

    i += 1
    with tabs[i]:
        st.code(inspect.getsource(czsc_strategy))


if st.session_state.run:
    trader = CzscTrader(bg=bg, positions=deepcopy(tactic.positions), signals_config=deepcopy(tactic.signals_config))
    bars1 = bars_right[0:st.session_state.bar_edt_index].copy()
    while bars1:
        bar_ = bars1.pop(0)
        trader.on_bar(bar_)

    bars2 = bars_right[st.session_state.bar_edt_index:].copy()
    with st.empty():
        while bars2:
            bar_ = bars2.pop(0)
            trader.on_bar(bar_)
            show_trader(trader)
            st.session_state.bar_edt_index += 1

else:
    trader = CzscTrader(bg=bg, positions=deepcopy(tactic.positions), signals_config=deepcopy(tactic.signals_config))
    bars2 = bars_right[0:st.session_state.bar_edt_index].copy()
    with st.empty():
        while bars2:
            bar_ = bars2.pop(0)
            trader.on_bar(bar_)
        show_trader(trader)



from loguru import logger
from czsc import CzscSignals
from collections import OrderedDict
from czsc.utils import create_single_signal


def cxt_intraday_V230701(cat: CzscSignals, **kwargs) -> OrderedDict:
    """每日走势分类

    参数模板："{freq1}#{freq2}_D{di}日_走势分类V230701"

    **信号逻辑：**

    参见博客：https://blog.sina.com.cn/s/blog_486e105c010009uy.html

    **信号列表：**

    - Signal('30分钟#日线_D2日_走势分类V230701_强平衡市_任意_任意_0')
    - Signal('30分钟#日线_D2日_走势分类V230701_弱平衡市_任意_任意_0')
    - Signal('30分钟#日线_D2日_走势分类V230701_双中枢下跌_任意_任意_0')
    - Signal('30分钟#日线_D2日_走势分类V230701_转折平衡市_任意_任意_0')
    - Signal('30分钟#日线_D2日_走势分类V230701_双中枢上涨_任意_任意_0')

    :param c: CZSC对象
    :return: 信号识别结果
    """
    di = int(kwargs.get("di", 2))
    freq1 = kwargs.get("freq1", "30分钟")
    freq2 = kwargs.get("freq2", "日线")
    assert freq1 == '30分钟', 'freq1必须为30分钟'
    assert freq2 == '日线', 'freq2必须为日线'

    assert 21 > di > 0, "di必须为大于0小于21的整数，暂不支持当日走势分类"
    k1, k2, k3 = f"{freq1}#{freq2}_D{di}日_走势分类V230701".split('_')
    v1 = "其他"
    if '30分钟' not in cat.kas.keys() or '日线' not in cat.kas.keys():
        logger.warning(f"缺少30分钟或日线K线数据，无法计算当日走势分类, {cat.kas.keys()}")
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)
    
    c1 = cat.kas[freq1]
    c2 = cat.kas[freq2]
    day = c2.bars_raw[-di].dt.date()
    bars = [x for x in c1.bars_raw if x.dt.date() == day]
    assert len(bars) <= 8, f"仅适用于A股市场，日内有8根30分钟K线的情况, {len(bars)}, {day}, {bars}"
    if len(bars) <= 4:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    zs_list = []
    for b1, b2, b3 in zip(bars[:-2], bars[1:-1], bars[2:]):
        if min(b1.high, b2.high, b3.high) > max(b1.low, b2.low, b3.low):
            zs_list.append([b1, b2, b3])
    
    _dir = "上涨" if bars[-1].close > bars[0].open else "下跌"

    if not zs_list:
        v1 = f"无中枢{_dir}"
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)
    
    # 双中枢的情况，有一根K线的 high low 与前后两个中枢没有重叠
    if len(zs_list) >= 2:
        zs1, zs2 = zs_list[0], zs_list[-1]
        zs1_high, zs1_low = max([x.high for x in zs1]), min([x.low for x in zs1])
        zs2_high, zs2_low = max([x.high for x in zs2]), min([x.low for x in zs2])
        if _dir == "上涨" and zs1_high < zs2_low:
            v1 = f"双中枢{_dir}"
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)
        
        if _dir == "下跌" and zs1_low > zs2_high:
            v1 = f"双中枢{_dir}"
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)
        
    # 单中枢的情况，前三根K线出现高点：弱平衡市，前三根K线出现低点：强平衡市，否则：转折平衡市
    high_first = max(bars[0].high, bars[1].high, bars[2].high) == max([x.high for x in bars])
    low_first = min(bars[0].low, bars[1].low, bars[2].low) == min([x.low for x in bars])
    if high_first and not low_first:
        v1 = "弱平衡市"
    elif low_first and not high_first:
        v1 = "强平衡市"
    else:
        v1 = "转折平衡市"
        
    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)



def check():
    from czsc.connectors import research
    from czsc.traders.base import check_signals_acc

    symbols = research.get_symbols('A股主要指数')
    bars = research.get_raw_bars(symbols[0], '15分钟', '20181101', '20210101', fq='前复权')

    signals_config = [{'name': cxt_intraday_V230701, 'di': 1, 'freq1': '30分钟', 'freq2': '日线'}]
    check_signals_acc(bars, signals_config=signals_config, height='780px', delta_days=0) # type: ignore


if __name__ == '__main__':
    check()
